"""Basic python utility functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/utilities/basic.ipynb.

# %% auto 0
__all__ = ['get_item', 'filter_args', 'filter_out_args']

# %% ../../nbs/api/utilities/basic.ipynb
import inspect
from typing import *

# %% ../../nbs/api/utilities/basic.ipynb
def get_item(
    l:Sequence,           # Sequence to get item of
    i:int,                # Index of item to get
    default:Optional=None # Default value to return if sequence does not contain item at index
)->Any:                   # Item or default value
    """Return item at index `i` from sequence `l` or `default` if item does not exist."""
    try: return l[i]
    except IndexError: return default

# %% ../../nbs/api/utilities/basic.ipynb
def filter_args(
    fnc:Callable, # Function to test keyword arguments against
    **kwargs      # Keyword arguments to test
)->dict:          # Keyword arguments defined in signature of `fnc` 
    """Return only keyword arguments defined in signature of function `fnc`."""
    return {k: v for k, v in kwargs.items() if k in inspect.signature(fnc).parameters}

# %% ../../nbs/api/utilities/basic.ipynb
def filter_out_args(
    fnc:Callable, # Function to test keyword arguments against
    **kwargs      # Keyword arguments to test
)->dict:          # Keyword arguments not defined in signature of `fnc` 
    """Return only keyword arguments not defined in signature of function `fnc`."""
    return {k: v for k, v in kwargs.items() if k not in inspect.signature(fnc).parameters}



"""Extra functionality for [nbdev](https://nbdev.fast.ai/)"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/utilities/nbdev.ipynb.

# %% auto 0
__all__ = ['MarkdownTable', 'DictTbl', 'AttributeTbl']

# %% ../../nbs/api/utilities/nbdev.ipynb
import inspect
from fastcore.foundation import L

# %% ../../nbs/api/utilities/nbdev.ipynb
def _split_at(s:str, *args, strip=False):
    result = [s]
    for arg in args:
        strs = result.pop().split(arg)
        if len(strs) == 2: result += strs
        else: return None
    if strip: result = [s.strip(strip) for s in result]
    return result        

# %% ../../nbs/api/utilities/nbdev.ipynb
def _bold(s): return f'**{s}**' if s.strip() else s

# %% ../../nbs/api/utilities/nbdev.ipynb
def _list2row(l:list): return '| '+' | '.join(l) + ' |'

# %% ../../nbs/api/utilities/nbdev.ipynb
class MarkdownTable:
    
    def __init__(self, data, headers):
        self.data = data
        self.headers = headers
    
    @property
    def _columns(self):
        return self._map

    @property
    def _hdr_list(self): return [_bold(l) for l in self.headers]

    @property
    def hdr_str(self):
        "The markdown string for the header portion of the table"
        md = _list2row(self._hdr_list)
        return md + '\n' + _list2row(['-' * len(l) for l in self._hdr_list])

    @property
    def params_str(self):
        "The markdown string for the parameters portion of the table."
        return '\n'.join(L(self.data).map(_list2row))

    def _repr_markdown_(self):
        _tbl = [self.hdr_str, self.params_str]
        return '\n'.join(_tbl)

    def __eq__(self,other): return self.__str__() == str(other).strip()

    __str__ = _repr_markdown_

# %% ../../nbs/api/utilities/nbdev.ipynb
class DictTbl(MarkdownTable):
    def __init__(self, obj):
        rows = []
        for s in inspect.getsource(obj).split('\n'):
            line = L(_split_at(s, ":", "#", strip=' '))
            if len(line) == 3 and line[0][0] in "\'\"": rows.append(s.strip('\"\'') for s in line[[0, -1]])
        super().__init__(rows, [' ', 'Description'])

# %% ../../nbs/api/utilities/nbdev.ipynb
class AttributeTbl(MarkdownTable):
    def __init__(self, obj):
        rows = []
        for s in inspect.getsource(obj.__init__).split('\n'):
            line = L(_split_at(s, " = ", " # ", strip=' '))
            if len(line) == 3 and line[0].startswith('self.'): rows.append([line[0][5:], line[2]])
        super().__init__(rows, ['Attribute', 'Description'])



"""Extended functionality for [pyMOR](https://pymor.org/) operators"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/pymor/operators.ipynb.

# %% ../../nbs/api/pymor/operators.ipynb 4
from __future__ import annotations
from pathlib import Path
import importlib
from numbers import Number
import math
from functools import reduce

from fastcore.basics import patch
from fastcore.meta import delegates
from pydash import unzip

import numpy as np
from numpy import array, ndarray
from scipy.sparse import issparse, sparray
from sparse import SparseArray
import sympy as sy
import pandas as pd
from xarray import DataArray, Coordinates

from pymor.basic import LincombOperator, NumpyMatrixOperator, ZeroOperator, ExpressionParameterFunctional, NumpyVectorSpace, Mu, IdentityOperator, Parameters
from pymor.models.interface import Model
from pymor.algorithms.to_matrix import to_matrix, ToMatrixRules
from pymor.algorithms.simplify import ExpandRules, expand, ContractRules, contract
from pymor.algorithms.rules import match_class, RuleTable
from pymor.algorithms.lincomb import assemble_lincomb
from pymor.operators.interface import Operator
from pymor.core.base import abstractmethod
from pymor.parameters.functionals import ParameterFunctional

from ..utilities.xarray import *
from ..utilities.sparse import sparse, sparse2d, kron, sparse2d_rand
from .parameters import *
from .vectorarrays import *

# %% auto 0
__all__ = ['operator_html_tree', 'coordinates_to_list', 'XarrayMatrixOperator', 'contract_matvec', 'dims_matvec', 'replace_item',
           'densify', 'XarrayFunctionalOperator', 'SumOperator', 'ScaleOperator', 'ProductOperator']

# %% ../../nbs/api/pymor/operators.ipynb
LincombOperator.suboperator_symbol = '+'

# %% ../../nbs/api/pymor/operators.ipynb
def _mapping_str(op:Operator):
    try:
        return f"{op.source._repr_html_(show_details=False)} → {op.range._repr_html_(show_details=False)}"
    except AttributeError:
        return f"{str(op.source)} → {str(op.range)}"

# %% ../../nbs/api/pymor/operators.ipynb
def _operator_group_tag(
    level:int, 
    cls:str, 
    mapping:str, 
    name:str|None=None,
    parameters:str|None=None,
    contents:str|None=None,
    suboperators:list(str)|None=None, 
    symbol:str|None=None, 
    opened:bool=False
) -> str:
    if suboperators or contents:
        return f'''
            <li>
                <details {'open' if opened else 'closed'}>
                    <summary>
                        <ul class="operator">
                            {'<li>' + name + '</li>' if name else ''}
                            <li>{cls}</li>
                            <li>{mapping}</li>
                            {'<li>' + parameters + '</li>' if parameters else ''}
                        </ul>
                    </summary>
                    <ul class="operator-group operator-group-{'even' if not (level % 2) else 'odd'}">
                        {contents or symbol.join(suboperators)}
                    </ul>
                </details>
            </li>
        '''
    else:
        return f'''
            <li>
                <ul class="operator">
                    {'<li>' + name + '</li>' if name else ''}
                    <li>{cls}</li>
                    <li>{mapping}</li>
                    {'<li>' + parameters + '</li>' if parameters else ''}
                </ul>
            </li>
        '''

# %% ../../nbs/api/pymor/operators.ipynb
def _operator_html_node(op:Operator, level:int=0, opened:bool=False) -> str:
    cls = op.__class__.__name__
    name = op.name if op.name != cls else None
    mapping = _mapping_str(op)
    parameters = str(list(Parameters.of(op)))
    contents = None
    if hasattr(op, 'operators'):
        suboperators = [_operator_html_node(o, level=level + 1, opened=opened) for o in op.operators]
        symbol = op.suboperator_symbol
        return _operator_group_tag(level, cls, mapping, name, parameters, contents, suboperators, symbol, opened)
    if hasattr(op, 'matrix'):
        contents = op.matrix._repr_html_() if hasattr(op.matrix, '_repr_html_') else str(op.matrix)
        return _operator_group_tag(level, cls, mapping, name, parameters, contents)
    return _operator_group_tag(level, cls, mapping, name, parameters)

# %% ../../nbs/api/pymor/operators.ipynb
def operator_html_tree(op:Operator, opened:bool=False) -> str:
    style = """
    <style>
        .operator-group summary {
            list-style: none;
            display: flex;
            gap: 0.25em;
            align-items: start;
            cursor: pointer;
        }
        .operator-group summary::before {
            content: '▶︎';
            font-size: 75%;
            align-items: center;
            width: 1em;
            flex-shrink: 0;
            transition: 0.3s;
        }
        .operator-group details[open] > summary::before {
            transform: rotate(90deg);
            transform-origin: 40% 45% 0;
        }
        .operator-group-odd > li {
            background-color: #f0f0f0;;
        }
        .operator-group-even > li {
            background-color: #fbfbfb;
        }
        .operator-group {
            display: flex;
            gap: .4em;
            align-items: center;
            padding: 0px;
            margin-left: -2em;
            margin-top: .25em;
        }
        .operator-group > li {
            list-style-type: none;
            border: 1px solid black;
            border-radius: .25em;
            padding: .25em;
        }
        .operator {
            margin: 0;
            padding: 0;
            margin-top: -.2em;
            margin-left: -2em;
        }
        .operator > li {
            list-style-type: none;
            border: none;
            padding: 0;
            line-height: 1.25em;
        }
    </style>
    """
    return style + f'<ul class="operator-group operator-group-odd">{_operator_html_node(op, opened=opened)}</ul>'

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def _repr_html_(self:Operator):
    return operator_html_tree(self)

# %% ../../nbs/api/pymor/operators.ipynb
class XarrayMatrixBasedOperator(Operator):
    """Base class for operators that assemble into an `XarrayMatrixOperator`."""

    linear = True

    @property
    def H(self):
        if not self.parametric:
            return self.assemble().H
        else:
            return super().H

    @abstractmethod
    def assemble(self, mu=None): pass

    def apply(self, U, mu=None):
        return self.range.from_xarray(self.assemble(mu).matrix.dot(U.impl._array))

    def apply_adjoint(self, V, mu=None):
        return self.assemble(mu).apply_adjoint(V)

    def as_range_array(self, mu=None):
        matrix = unlabel_dims(self.assemble(mu).matrix.squeeze())
        extended_coords = {d: matrix.coords[d] for d in matrix.dims if d not in self.range.dims}
        return self.range.from_numpy(self.matrix.data, extended_coords)

    def as_source_array(self, mu=None):
        matrix = unlabel_dims(self.assemble(mu).matrix.squeeze())
        extended_coords = {d: matrix.coords[d] for d in matrix.dims if d not in self.source.dims}
        return self.source.from_numpy(self.matrix.data, extended_coords).conj()

    @property
    def display_name(self):
        return self.name if self.name != self.__class__.__name__ else 'X'

    def __str__(self):
        return f"{self.display_name}{_mapping_str(self)}"

# %% ../../nbs/api/pymor/operators.ipynb
def _ndims(matrix):
    if isinstance(matrix, DataArray):
        return len(matrix.dims)
    else:
        return matrix.ndim    

# %% ../../nbs/api/pymor/operators.ipynb
def _label_dims_with(obj, lbl):
    if isinstance(obj, XarrayVectorSpace):
        return obj.with_(coord_data = _label_dims_with(obj.coord_data, lbl))
    if isinstance(obj, list | tuple): 
        return [d + lbl for d in obj]
    if isinstance(obj, dict):
        return {k + lbl: v for k,v in obj.items()}

# %% ../../nbs/api/pymor/operators.ipynb
def _label_dims(obj, src_dims, rng_dims):
    if isinstance(obj, XarrayVectorSpace):
        coord_data = {_label_dims(k, src_dims, rng_dims): v for k,v in obj.coord_data.items()}
        return obj.with_(coord_data=coord_data)
    if isinstance(src_dims, str): src_dims = [src_dims]
    if isinstance(rng_dims, str): rng_dims = [rng_dims]
    labled_src_dims = _label_dims_with(src_dims, Lbl.SRC)
    labled_rng_dims = _label_dims_with(rng_dims, Lbl.RNG)
    rename_dict = dict(zip(src_dims, labled_src_dims)) | dict(zip(rng_dims, labled_rng_dims))
    match obj:
        case str():
            return rename_dict.get(obj, obj)
        case list() | tuple():
            return [rename_dict.get(dim, dim) for dim in obj]
        case DataArray():
            return obj.rename(rename_dict)
        case Coordinates():
            return Coordinates({rename_dict.get(k, k): v.data for k,v in obj.items()})

# %% ../../nbs/api/pymor/operators.ipynb
def _swap_dim_labels(da):
    src_dims = [d for d in da.dims if d.endswith(Lbl.SRC)]
    rng_dims = [d for d in da.dims if d.endswith(Lbl.RNG)]
    rename_dict = {d: d.replace(Lbl.SRC, Lbl.RNG) for d in src_dims} | {d: d.replace(Lbl.RNG, Lbl.SRC) for d in rng_dims}
    return da.rename(rename_dict)

# %% ../../nbs/api/pymor/operators.ipynb
def _is_labeled_da(da):
    if not isinstance(da, DataArray): return False
    return any(d.endswith(Lbl.SRC) or d.endswith(Lbl.RNG) for d in da.dims)

# %% ../../nbs/api/pymor/operators.ipynb
def coordinates_to_list(coords, dims):
    return [(d, coords[d].data, coords[d].attrs) for d in dims]

# %% ../../nbs/api/pymor/operators.ipynb
def _get_space_dims(dims):
    if dims is None: return None, None
    if isinstance(dims, dict): dims = Coordinates(dims)
    if isinstance(dims, Coordinates): return list(dims), dims
    if isinstance(dims, str): dims = [dims]
    if isinstance(dims, list | tuple): return dims, None
    if hasattr(dims, 'dims'): return dims.dims, dims.coords
    raise ValueError("Source or range specification not valid")

# %% ../../nbs/api/pymor/operators.ipynb
class XarrayMatrixOperator(XarrayMatrixBasedOperator):
    """An `Operator` backed by an xarray `DataArray`."""

    def __init__(
        self, 
        matrix:DataArray|ndarray|SparseArray|sparray, # N-dimensional matrix
        source:XarrayVectorSpace|DataArray|Coordinates|dict|list|str|None=None, # Source vector space or coordinates
        range:XarrayVectorSpace|DataArray|Coordinates|dict|list|str|None=None, # Range vector space or coordinates
        solver_options:dict=None, # Options for matrix solver
        name:str=None # Operator name
    ):
        if issparse(matrix): matrix = sparse(matrix)
        if not isinstance(matrix, DataArray):
            if not isinstance(source, XarrayVectorSpace): 
                if isinstance(source, str): source = [source]
                if source is not None: 
                    source = validate_coord_data(source, matrix.shape[-len(source):])
                source = XarrayVectorSpace(source)
            if not isinstance(range, XarrayVectorSpace):
                if isinstance(range, str): range = [range]
                if range is not None:
                    range = validate_coord_data(range, matrix.shape[:len(range)])
                range = XarrayVectorSpace(range)
            labeled_source = _label_dims_with(source, Lbl.SRC)
            labeled_range = _label_dims_with(range, Lbl.RNG)
            matrix = DataArray(
                matrix, 
                dims=labeled_range.dims + labeled_source.dims,
                coords=make_coordinates(labeled_range.coord_data | labeled_source.coord_data),
                name=name
            )
        else:
            if not _is_labeled_da(matrix): matrix = _label_dims(matrix, source, range)
            coords = validate_coord_data(matrix)
            labeled_source = XarrayVectorSpace({k:v for k,v in coords.items() if k.endswith(Lbl.SRC)})
            labeled_range = XarrayVectorSpace({k:v for k,v in coords.items() if k.endswith(Lbl.RNG)})
            source = unlabel_dims(labeled_source)
            range = unlabel_dims(labeled_range)
        if name is None: name = matrix.name
        self.__auto_init(locals())
        self.labeled_source = labeled_source
        self.labeled_range = labeled_range

    def assemble(self, mu=None): return self

    @property
    def size(self):
        return self.matrix.size
        
    @property
    def sparse(self):
        return isinstance(self.matrix.data, SparseArray)

# %% ../../nbs/api/pymor/operators.ipynb
@patch(as_prop=True)
def H(self:XarrayMatrixOperator):
    """The adjoint operator."""
    adjoint_matrix = self.matrix if np.isrealobj(self.matrix) else self.matrix.conj()           
    options = {'inverse': self.solver_options.get('inverse_adjoint'),
               'inverse_adjoint': self.solver_options.get('inverse')} if self.solver_options else None
    return self.with_(matrix=_swap_dim_labels(adjoint_matrix), solver_options=options, name=self.name + '_adjoint')

# %% ../../nbs/api/pymor/operators.ipynb
def contract_matvec(a, b, axes):
    """Contract one axis of the first (2D) array with one axis of the second (ND) array."""
    out = np.tensordot(a, b, axes)
    if a.ndim == 1: return out
    return np.moveaxis(out, 0, axes[1])

# %% ../../nbs/api/pymor/operators.ipynb
def dims_matvec(a, b):
    """Contract one axis of the first (2D) array with the axis of the second (ND) array of the same name."""
    a, adims = a
    b, bdims = b
    ax = [[adims.index(d), i] for i,d in enumerate(bdims) if d in adims][0]
    newdims = list(bdims)
    if len(adims) == 1:
        newdims.pop(ax[1])
    else:
        newdims[ax[1]] = adims[1 - ax[0]]
    return contract_matvec(a, b, ax), newdims

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def __truediv__(self:XarrayVectorSpace, other):
    if not isinstance(other, XarrayVectorSpace): raise NotImplementedError
    return self.with_(coords={k: v for k,v in self.coords.items() if k not in other.coords})

# %% ../../nbs/api/pymor/operators.ipynb
def replace_item(dct, old_key, new_pair):
    return dict(new_pair if k == old_key else (k, v) for k,v in dct.items())

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def range_under(self:XarrayVectorSpace, op:XarrayMatrixOperator):
    """Find the space that this space is mapped to by an XarrayMatrixOperator. Assumes that the operator has one-dimensional source and range."""
    source_dim = op.source.dims[0]
    if not op.range.dims:
        coords = {k:v for k,v in self.coord_data.items() if k != source_dim}
    else:
        range_dim = op.range.dims[0]
        range_coord = op.range.coord_data[range_dim]
        coords = replace_item(self.coord_data, source_dim, (range_dim, range_coord))
    return self.with_(coord_data=coords)

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def apply(
    self:XarrayMatrixOperator, 
    U, # `XarrayVectorArray` of vectors to which the operator is applied
    mu=None, # The parameter values for which to evaluate the operator
    method='numpy',
)->XarrayVectorArray: # `XarrayVectorArray` in the range `XarrayVectorSpace`
    """Apply the operator to an `XarrayVectorArray` in the source vector space."""
    assert U in self.source
    if method == 'numpy':
        src_dims = _label_dims(U.array.dims, U.dims, [])
        out, out_dims = dims_matvec((self.matrix.data, self.matrix.dims), (U.array.data, src_dims))
        out = U.space.range_under(self).from_numpy(out, extended_coord_data=U.extended_coord_data)
    else:
        out = U.space.range_under(self).from_xarray(unlabel_dims(self.matrix.dot(U.array_labeled_as_source())))
    if self.name != self.__class__.__name__: out.name = self.name
    return out

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def apply_adjoint(
    self:XarrayMatrixOperator, 
    V, # `XarrayVectorArray` of vectors to which the operator is applied
    mu=None # The parameter values for which to evaluate the operator
)->XarrayVectorArray: # `XarrayVectorArray` in the source `XarrayVectorSpace`
    """Apply the adjoint of the operator to an `XarrayVectorArray` in the range vector space."""
    return self.H.apply(V, mu=mu)

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def _assemble_lincomb(self:XarrayMatrixOperator, operators, coefficients, identity_shift=0., name=None):
    if not all(isinstance(op, XarrayMatrixOperator) for op in operators):
        return None

    common_mat_dtype = reduce(np.promote_types,
                              (op.matrix.dtype for op in operators if hasattr(op, 'matrix')))
    common_coef_dtype = reduce(np.promote_types, (type(c) for c in coefficients + [identity_shift]))
    common_dtype = np.promote_types(common_mat_dtype, common_coef_dtype)

    if coefficients[0] == 1:
        matrix = operators[0].matrix.astype(common_dtype)
    else:
        matrix = operators[0].matrix * coefficients[0]
        if matrix.dtype != common_dtype:
            matrix = matrix.astype(common_dtype)

    for op, c in zip(operators[1:], coefficients[1:], strict=True):
        if c == 1:
            try:
                matrix += op.matrix
            except NotImplementedError:
                matrix = matrix + op.matrix
        elif c == -1:
            try:
                matrix -= op.matrix
            except NotImplementedError:
                matrix = matrix - op.matrix
        else:
            try:
                matrix += (op.matrix * c)
            except NotImplementedError:
                matrix = matrix + (op.matrix * c)

    if identity_shift: raise NotImplementedError

    return XarrayMatrixOperator(matrix, source=self.source, range=self.range)

# %% ../../nbs/api/pymor/operators.ipynb
@match_class(XarrayMatrixOperator)
def to_matrix_XarrayMatrixOperator(self, op):
    """Return the operator as a 2D matrix with stacked range and source dimensions in either `ndarray` or `sparray` format."""
    data = op.matrix.stack(_range=op.labeled_range.coords.dims, _source=op.labeled_source.coords.dims).data
    if isinstance(data, SparseArray): return sparse2d(data)
    return data
    
ToMatrixRules.insert_rule(-1, to_matrix_XarrayMatrixOperator)

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def to_numpy(self:XarrayMatrixOperator):
    """Return the operator in dense `ndarray` format."""
    return self.matrix.to_numpy()

# %% ../../nbs/api/pymor/operators.ipynb
class DensifyRules(RuleTable):
    """|RuleTable| for the :func:`densify` algorithm."""

    def __init__(self):
        super().__init__(use_caching=True)

    @match_class(XarrayMatrixOperator)
    def action_XarrayMatrixOperator(self, op):
        op = self.replace_children(op)
        if op.sparse:
            matrix = op.matrix.copy()
            matrix.data = matrix.data.todense()
            op = op.with_(matrix=matrix)
        return op
    
    @match_class(Model, Operator)
    def action_recurse(self, op):
        return self.replace_children(op)

# %% ../../nbs/api/pymor/operators.ipynb
def densify(
    obj:Model|Operator # Object to densify
)->Model|Operator: # Densified object
    """Convert sparse operators to dense."""
    return DensifyRules().apply(obj)

# %% ../../nbs/api/pymor/operators.ipynb
class XarrayFunctionalOperator(XarrayMatrixBasedOperator):
    """An `Operator` described by a `ParameterFunctional` that assembles to a `XarrayMatrixOperator`."""
    
    def __init__(self, functional:ParameterFunctional, range, source):
        self.__auto_init(locals())

    def assemble(self, mu=None): 
        return XarrayMatrixOperator(self.functional.evaluate(mu), range=self.range.dims, source=self.source.dims)

    def apply(self, U:VectorArray, mu=None):
        return self.assemble(mu).apply(U)

    @property
    def size(self):
        return self.matrix.size

# %% ../../nbs/api/pymor/operators.ipynb
class SumOperator(Operator):
    """An `Operator` that sums over one or more dimensions of a `XarrayVectorSpace`."""
    
    linear = True
    
    def __init__(
        self, 
        source_coords:Coordinates|dict, # Coordinates to sum over 
        name=None # Operator name
    ):
        self.__auto_init(locals())
        self.source = XarrayVectorSpace(source_coords)
        self.range = XarrayVectorSpace("Sum")

    def assemble(self, mu=None): return self

    def apply(self, U, mu=None):
        U = U.space.range_under(self).from_xarray(U.array.sum(self.source.dims).expand_dims('Sum'))
        if self.name != self.__class__.__name__: U.name = self.name
        return U

    def __mul__(self, other):
        if isinstance(other, Operator):
            return ProductOperator([self, other])
        return super().__mul__(other)

    def __rmul__(self, other):
        if isinstance(other, Operator):
            return ProductOperator([other, self])
        return super().__mul__(other)

    def __str__(self):
        return f"{self.name if self.name != self.__class__.__name__ else '∑'}{_mapping_str(self)}"

# %% ../../nbs/api/pymor/operators.ipynb
class ScaleOperator(Operator):
    """A scaling operator for `XarrayVectorSpace`s."""
    
    def __init__(self, array, space=None, name=None):
        if not isinstance(array, DataArray): array = DataArray(array, coords=space.coords, name=name)
        if space is None: space = XarrayVectorSpace(array)
        self.__auto_init(locals())
        self.source = space
        self.range = space

    def assemble(self, mu=None): return self

    def apply(self, U, mu=None):
        U = U * self.array
        if self.name != self.__class__.__name__: U.name = self.name
        return U

    @property
    def display_name(self):
        return self.name if self.name != self.__class__.__name__ else 'S'
    
    def __mul__(self, other):
        if isinstance(other, Operator):
            return ProductOperator([self, other])
        return super().__mul__(other)

    def __rmul__(self, other):
        if isinstance(other, Operator):
            return ProductOperator([other, self])
        return super().__mul__(other)

    def __str__(self):
        return f"{self.display_name}{_mapping_str(self)}"

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def __init__(self:LincombOperator, operators, coefficients, solver=None, name=None):
    assert len(operators) > 0
    assert len(operators) == len(coefficients)
    assert all(isinstance(op, Operator) for op in operators)
    assert all(isinstance(c, (ParameterFunctional, Number)) for c in coefficients)
    # assert all(op.source == operators[0].source for op in operators[1:])
    # assert all(op.range == operators[0].range for op in operators[1:])
    operators = tuple(operators)
    coefficients = tuple(coefficients)

    self.operators = operators
    self.coefficients = coefficients
    self.solver = solver
    self.name = name
    try:
        self.source = max(self.operators, key=lambda o: o.source.ndim).source
        self.range = max(self.operators, key=lambda o: o.range.ndim).range
    except AttributeError:
        self.source = self.operators[0].source
        self.range = self.operators[0].range
    self.linear = all(op.linear for op in operators)


# %% ../../nbs/api/pymor/operators.ipynb
def _numpy_zero_operator(shape):
    return ZeroOperator(*[NumpyVectorSpace(dim) for dim in shape])

# %% ../../nbs/api/pymor/operators.ipynb
def _read_sparse_dataarray(
    file_name:str|Path, # File name
)->DataArray:
    if not Path(file_name).suffix: file_name = Path(file_name).with_suffix('.mtxn')
    header = []
    with open(file_name, "r") as reader:
        while True:
            line = next(reader)
            if line[0] != "%": break
            header.append(line.strip('%\n '))
    matrix = pd.read_csv(
        file_name,
        sep=r'\s+',
        header=None,
        skiprows=len(header),
        index_col=None,
        engine="c",
        comment='%'
    )
    name = header[0]
    parameters = header[1].split('\t')
    shape = [int(s) for s in header[2].split()]
    coords = {dim: index.split('\t') for dim, index in zip(header[3::2], header[4::2])}
    matrix = sparse([matrix.iloc[:, :-1].T, matrix.iloc[:, -1]], shape=shape)
    return DataArray(matrix, coords, name=name, attrs={'parameters': parameters})

# %% ../../nbs/api/pymor/operators.ipynb
@patch(cls_method=True)
@delegates(LincombOperator.__init__)
def from_file(
    cls:LincombOperator, 
    file_name:str|Path, # File name
    **kwargs # Additional arguments are passed to `LincombOperator`.
):
    """Read a `LincombOperator` (list of sparse arrays and corresponding `ExpressionParameterFunctionals`) from a .mtxn file."""
    da = _read_sparse_dataarray(file_name)
    coefficients = [
        ExpressionParameterFunctional(s, {p: 1 for p in da.parameters if p in s}) if da.parameters != [''] else 1
        for s in da.coords[da.dims[0]].data
    ]
    del da.attrs['parameters']
    # return da
    operators = [XarrayMatrixOperator(d.drop_vars(da.dims[0])) for d in da]    
    return LincombOperator(operators, coefficients, **kwargs)

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def partial_evaluate(
    self:ExpressionParameterFunctional, 
    mu:dict|Mu # Parameter values 
)->float|ExpressionParameterFunctional:
    """Substitute parameter values, returning a new `ExpressionParameterFunctional` if expression does not evaluate to a number."""
    mu = {k: v[0] for k, v in mu.items()}
    expr = sy.sympify(self.expression).subs(mu)
    unevaluated = [str(s) for s in expr.free_symbols]
    if expr.is_number:
        return float(expr)
    return ExpressionParameterFunctional(
        str(expr), 
        {k: v for k, v in self.parameters.items() if k in unevaluated}
    )

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def partial_evaluate_coefficients(
    self:LincombOperator, 
    mu:Mu # Parameter values to substitute
):
    """Substitute parameter values into linear coefficients, returning a new `ExpressionParameterFunctional` if expression does not evaluate to a number."""
    result = []
    for c in self.coefficients:
        try:
            result.append(c.evaluate(mu))
        except AttributeError:
            result.append(c)
        except AssertionError:
            try:
                result.append(c.partial_evaluate(mu))
            except Exception as e:
                result.append(c)
    return result

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def partial_assemble(
    self:LincombOperator, 
    mu=None # Parameter values to substitute
)->Operator:
    """Substitute parameter values into the linear coefficients, returning a new operator with fewer (or no) parameters."""
    operators = tuple(op.assemble(mu) for op in self.operators)
    coefficients = self.partial_evaluate_coefficients(mu)
    evaluated = [[], []]
    unevaluated = [[], []]
    for o, c in zip(operators, coefficients):
        if isinstance(c, Number):
            if c:
                evaluated[0].append(o)
                evaluated[1].append(c)
        else:
            unevaluated[0].append(o)
            unevaluated[1].append(c)
    op = []
    if len(evaluated[0]):
        op.append(assemble_lincomb(
            *evaluated, 
            # solver_options=self.solver_options,
            name=self.name + '_partially_assembled'
        ))
    if len(unevaluated[0]):
        op.append(LincombOperator(*unevaluated))
    if len(op) == 0:
        op = [ZeroOperator(range=self.range, source=self.source)]
    op = np.sum(op)
    # To avoid infinite recursions, only use the result if at least one of the following
    # is true:
    #   - The operator is parametric, so the the result of assemble *must* be a different,
    #     non-parametric operator.
    #   - One of self.operators changed by calling 'assemble' on it.
    #   - The result of assemble_lincomb is of a different type than the original operator.
    #   - assemble_lincomb could simplify the list of assembled operators,
    #     which we define to be the case when the number of operators has ben reduced.
    if (self.parametric
            or operators != self.operators  # for this comparison to work self.operators always has to be a tuple!
            or type(op) != type(self)
            or len(op.operators) < len(operators)):
        return op
    else:
        return self

# %% ../../nbs/api/pymor/operators.ipynb
@patch(as_prop=True)
def terms(self:LincombOperator):
    return [c * o for c, o in zip(self.coefficients, self.operators)]

# %% ../../nbs/api/pymor/operators.ipynb
class ProductOperator(Operator):
    """An Operator given by the direct product of operators `operators`."""
    
    suboperator_symbol = '⛒'
    
    def __init__(
        self, 
        operators, # Sequence of operators that each assemble to XarrayMatrixOperator
        name=None, # Optional name
    ):
        for op in operators:
            if not isinstance(op, Operator): raise ValueError(f'Input {op} should be an Operator.')
        self.source = np.prod([op.source for op in operators])
        self.range = np.prod([op.range for op in operators])
        self.__auto_init(locals())
        self.linear = True
        
    def apply(self, U, mu=None):
        pass

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def __mul__(self:Operator, other):
    if isinstance(other, Operator) and all(isinstance(space, XarrayVectorSpace) for space in [self.source, self.range, other.source, other.range]):
        return ProductOperator([self, other])
    if isinstance(other, (Number, ParameterFunctional)):
        if self.name != 'LincombOperator' or not isinstance(self, LincombOperator):
            return LincombOperator((self,), (other,))
        else:
            return self.with_(coefficients=tuple(c * other for c in self.coefficients))
    else:
        raise TypeError(f'Object {other} is not an Operator, Number, or ParameterFunctional.')

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def __str__(self:ProductOperator):
    return ' ⨂ '.join(str(o) for o in self.operators)

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def __str__(self:IdentityOperator):
    return f'I{_mapping_str(self)}'

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def apply(self:ProductOperator, U, mu=None):
    U = U.copy()
    for op in self.operators[::-1]:
        U = op.assemble(mu).apply(U)
    return U.space.range_under(self).from_xarray(U.array)

# %% ../../nbs/api/pymor/operators.ipynb
@patch(as_prop=True)
def H(self:ProductOperator):
    return self.with_(operators=[op.H for op in self.operators])

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def apply_adjoint(self:ProductOperator, V, mu=None):
    return self.H.apply(V, mu=mu)

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def assemble(self:ProductOperator, mu=None):
    operators = [op.assemble(mu) for op in self.operators]
    if any(isinstance(op, ZeroOperator) for op in operators): return ZeroOperator(self.range, self.source)
    if all(op == old_op for op, old_op in zip(operators, self.operators)): return self 
    return self.with_(operators=operators)

# %% ../../nbs/api/pymor/operators.ipynb
@match_class(ProductOperator)
def expand_ProductOperator(self, op):
    op = self.replace_children(op)

    # merge child ProductOperators
    if any(isinstance(o, ProductOperator) for o in op.operators):
        ops = []
        for o in op.operators:
            if isinstance(o, ProductOperator):
                ops.extend(o.operators)
            else:
                ops.append(o)
        op = op.with_(operators=ops)

    # expand products with LincombOperators
    if any(isinstance(o, LincombOperator) for o in op.operators):
        i = next(iter(i for i, o in enumerate(op.operators) if isinstance(o, LincombOperator)))
        left, right = op.operators[:i], op.operators[i+1:]
        ops = [ProductOperator([*left, o, *right]) for o in op.operators[i].operators]
        op = op.operators[i].with_(operators=ops)

        # there can still be LincombOperators within the summands so we recurse ..
        op = self.apply(op)
    return op

#| export
ExpandRules.insert_rule(-1, expand_ProductOperator)

# %% ../../nbs/api/pymor/operators.ipynb
@match_class(IdentityOperator)
def expand_IdentityOperator(self, op):
    if not isinstance(op.source, XarrayVectorSpace): return op
    return ProductOperator([IdentityOperator(XarrayVectorSpace({k:v})) for k,v in op.space.coord_data.items()])

#| export
ExpandRules.insert_rule(-1, expand_IdentityOperator)

# %% ../../nbs/api/pymor/operators.ipynb
@match_class(ProductOperator)
def contract_ProductOperator(self, op):
    op = self.replace_children(op)
    if any(isinstance(o, ZeroOperator) for o in op.operators): return ZeroOperator(op.range, op.source)
    xarray_operators = [o for o in op.operators if     isinstance(o, XarrayMatrixOperator)]
    other_operators  = [o for o in op.operators if not isinstance(o, XarrayMatrixOperator)]
    if xarray_operators:
        matrices, ranges, sources = unzip([(o.matrix, o.range, o.source) for o in xarray_operators])
        op = XarrayMatrixOperator(
            math.prod(matrices), 
            source=math.prod(sources, start=XarrayVectorSpace()), 
            range=math.prod(ranges, start=XarrayVectorSpace())
        )
    else:
        op = 1
    if other_operators:
        op *= math.prod(other_operators) 
    return op

ContractRules.insert_rule(-2, contract_ProductOperator)

# %% ../../nbs/api/pymor/operators.ipynb
@match_class(ProductOperator)
def to_matrix_ProductOperator(self, op):
    return kron(*[self.apply(o) for o in op.operators])
    
ToMatrixRules.insert_rule(-2, to_matrix_ProductOperator)

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def as_source_array(self:ProductOperator, mu=None):
    return contract(self.assemble(mu)).as_source_array()

# %% ../../nbs/api/pymor/operators.ipynb
@patch
def as_range_array(self:ProductOperator, mu=None):
    return contract(self.assemble(mu)).as_range_array()



"""Generic imports for all modules in PyLGS"""

# standard library
from typing import *
from numbers import Number, Integral
import functools
from functools import partial
import re
import string
import warnings
import glob
import logging
import os
import itertools as it
from pathlib import Path
import warnings
from collections import namedtuple

#jupyter
from ipywidgets import widgets
from IPython.display import HTML

# programming tools
from fastcore.basics import ifnone, zip_cycle, listify, flatten, first, patch, patch_to, store_attr
from fastcore.meta import delegates
from plum import dispatch
from pydash import py_, map_values, unzip
import wrapt

# scientific packages
import numpy as np
from numpy import array, ndarray
from pandas import DataFrame, Series
from einops import rearrange, reduce, repeat
import sympy as sy
from sympy.parsing.sympy_parser import parse_expr

#pint
# import pint
# from pint import Quantity
# u = pint.UnitRegistry(force_ndarray_like=True)
# u.default_format = "~P"
# pint.set_application_registry(u)

# scipy
import scipy as sp
import scipy.linalg as spl
from scipy.special import erf, erfinv

# pymor
import pymor
from pymor.basic import Mu, Parameters, ExpressionParameterFunctional, NumpyGenericOperator, NumpyMatrixOperator, LincombOperator, NumpyVectorSpace, StationaryModel, InstationaryModel, GenericParameterFunctional
from pymor.core.base import ImmutableObject
from pymor.parameters.base import ParametricObject
from pymor.algorithms.to_matrix import to_matrix
from pymor.algorithms.simplify import expand, contract

# xarray and friends
import xarray as xr
from xarray import DataArray, Dataset
# import pint_xarray
# import xrft as xf
xr.set_options(keep_attrs=True)

warnings.filterwarnings("ignore", category=FutureWarning)



class Dummy():
    pass


"""Extended functionality for [pyMOR](https://pymor.org/) parameters"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/pymor/parameters.ipynb.

# %% auto 0
__all__ = ['SimplifyFunctionalRules', 'simplify_functionals']

# %% ../../nbs/api/pymor/parameters.ipynb
from numbers import Number
from fastcore.basics import patch
import sympy as sy
from sympy.parsing.sympy_parser import parse_expr
from pymor.parameters.functionals import ProductParameterFunctional, ExpressionParameterFunctional, ParameterFunctional
from pymor.algorithms.rules import match_class, RuleTable, match_always
from pymor.basic import LincombOperator
from pymor.models.interface import Model
from pymor.operators.interface import Operator

# %% ../../nbs/api/pymor/parameters.ipynb
@patch
def __eq__(self:ExpressionParameterFunctional, other):
    return (
        isinstance(other, ExpressionParameterFunctional)
        and self.expression == other.expression
        and self.parameters == other.parameters
    )

# %% ../../nbs/api/pymor/parameters.ipynb
@patch
def __str__(self:ParameterFunctional):
    return f'f({", ".join(self.parameters)})'

# %% ../../nbs/api/pymor/parameters.ipynb
class SimplifyFunctionalRules(RuleTable):
    """|RuleTable| for the :func:`expand` algorithm."""

    def __init__(self):
        super().__init__(use_caching=True)

    @match_class(ProductParameterFunctional)
    def action_ProductParameterFunctional(self, functional):
        # merge child ProductParameterFunctional objects
        if any(isinstance(factor, ProductParameterFunctional) for factor in functional.factors):
            factors = []
            for factor in functional_factors:
                if isinstance(factor, ProductParameterFunctional):
                    factors.extend(self.apply(factor).factors)
                else:
                    factors.append(factor)
            functional = functional.with_(factors=factors)

        # multiply together numbers and ExpressionParameterFunctional objects
        if all(isinstance(factor, (ExpressionParameterFunctional, Number)) for factor in functional.factors):
            product = sy.prod([
                parse_expr(factor.expression) if isinstance(factor, ExpressionParameterFunctional) else factor 
                for factor in functional.factors
            ])
            if product.is_number:
                functional = product
            else:
                functional = ExpressionParameterFunctional(str(product), parameters=functional.parameters)
        
        return functional
    
    @match_class(LincombOperator)
    def action_LincombOperator(self, op):
        return op.with_(coefficients=[self.apply(c) for c in op.coefficients])

    @match_class(Model, Operator)
    def action_recurse(self, op):
        return self.replace_children(op)

    @match_always
    def action_generic(self, expr):
        return expr

# %% ../../nbs/api/pymor/parameters.ipynb
def simplify_functionals(obj):
    return SimplifyFunctionalRules().apply(obj)



"""Extended functionality for [pyMOR](https://pymor.org/) time steppers"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/pymor/timestepping.ipynb.

# %% auto 0
__all__ = ['init_progress_widget', 'tick', 'update_progress_widget', 'init_cvode', 'solve_cvode', 'AdamsTimeStepper',
           'BDFTimeStepper', 'StiffSwitchingTimeStepper']

# %% ../../nbs/api/pymor/timestepping.ipynb
import time

from fastcore.basics import patch
import ipywidgets as widgets

import numpy as np
from scikits.odes.sundials import cvode
import scipy.sparse as sps
import pymor
from pymor.algorithms.to_matrix import to_matrix
from pymor.algorithms.simplify import expand
from pymor.algorithms.timestepping import TimeStepper
from pymor.basic import LincombOperator

from pymor.algorithms.simplify import ExpandRules, expand, ContractRules, contract
from pymor.algorithms.rules import match_class, RuleTable


from ..utilities.sparse import sparse2d_identity, spilu, restrict_bandwidth
from .vectorarrays import *
from .operators import *
# from pymor.basic import *

# %% ../../nbs/api/pymor/timestepping.ipynb
try:
    @pymor.defaults('cvode_bdf_rtol', 'cvode_bdf_atol', 'cvode_bdf_max_steps', 'cvode_stiff_switching_t_switch')
    def cvode_solver_options(
        cvode_bdf_rtol=1e-3,
        cvode_bdf_atol=1e-6,
        cvode_bdf_max_steps=100000,
        cvode_bdf_inflection_times=None,
        cvode_stiff_switching_t_switch=1e-7
    ):
        opts = {
            'cvode_bdf': {'type': 'cvode_bdf', 'atol': cvode_bdf_atol, 'rtol': cvode_bdf_rtol, 'max_steps': cvode_bdf_max_steps, 'inflection_times': cvode_bdf_inflection_times},
            'cvode_stiff_switching': {'type': 'cvode_stiff_switching', 't_switch': cvode_stiff_switching_t_switch}
        }
        return opts
except ValueError: pass

# %% ../../nbs/api/pymor/timestepping.ipynb
@patch
def solve(self:TimeStepper, initial_time, end_time, initial_data, operator, rhs=None, mass=None, mu=None, num_values=None):
    """Apply time-stepper to the equation.

    The equation is of the form ::

        M(mu) * d_t u + A(u, mu, t) = F(mu, t),
                         u(mu, t_0) = u_0(mu).

    Parameters
    ----------
    initial_time
        The time at which to begin time-stepping.
    end_time
        The time until which to perform time-stepping.
    initial_data
        The solution vector at initial_timeinitial_time.
    operator
        The |Operator| A.
    rhs
        The right-hand side F (either |VectorArray| of length 1 or |Operator| with
        source.dim == 1source.dim == 1). If NoneNone, zero right-hand side is assumed.
    mass
        The |Operator| M. If NoneNone, the identity operator is assumed.
    mu
        |Parameter values| for which operatoroperator and rhsrhs are evaluated. The current
        time is added to mumu with key tt.
    num_values
        The number of returned vectors of the solution trajectory. If NoneNone, each
        intermediate vector that is calculated is returned.

    Returns
    -------
    |VectorArray| containing the solution trajectory.
    """
    try:
        num_time_steps = self.estimate_time_step_count(initial_time, end_time)
    except NotImplementedError:
        num_time_steps = 0
    iterator = self.iterate(initial_time, end_time, initial_data, operator, rhs=rhs, mass=mass, mu=mu,
                            num_values=num_values)
    iterator = list(iterator)
    if isinstance(iterator[0], XarrayVectorArray):
        return iterator[0]
    U = operator.source.empty(reserve=num_values if num_values else num_time_steps + 1)
    t = []
    for U_n, t_n in iterator:
        U.append(U_n)
        t.append(t_n)
    try: return operator.source.from_numpy(U.to_numpy(), l={"Time": t})
    except: return U

# %% ../../nbs/api/pymor/timestepping.ipynb
def init_progress_widget(initial_time, end_time):
    progress = widgets.FloatProgress(
        value=initial_time,
        min=initial_time,
        max=end_time,
        bar_style='info',
        orientation='horizontal'
    )
    step_size = widgets.Label()
    widget = widgets.HBox([progress, step_size])
    display(widget)
    return widget

# %% ../../nbs/api/pymor/timestepping.ipynb
def tick(): return time.time() * 1000

# %% ../../nbs/api/pymor/timestepping.ipynb
def update_progress_widget(widget, tock, t, last_t):
    newtick = tick()
    if newtick - tock < 10: return tock
    widget.children[0].value = t
    widget.children[1].value = f'{t - last_t:1.1e}'
    return newtick

# %% ../../nbs/api/pymor/timestepping.ipynb
def init_cvode(cvode_rhs, num_values):
    return cvode.CVODE(cvode_rhs, lmm_type='Adams', nonlinsolver='fixedpoint', max_steps=1000000, one_step_compute=num_values is None)

# %% ../../nbs/api/pymor/timestepping.ipynb
def solve_cvode(solver, t_list, initial_data):
    return solver.solve(t_list, initial_data.to_numpy().T[0])

# %% ../../nbs/api/pymor/timestepping.ipynb
class AdamsTimeStepper(TimeStepper):
    def __init__(self):
        self.__auto_init(locals())
        
    def iterate(self, initial_time, end_time, initial_data, operator, rhs=None, mass=None, mu=None, num_values=None):

        def cvode_rhs(t, y, ydot):
            y = operator.source.from_numpy(y)
            np.copyto(ydot, rhs.as_range_array().to_numpy().ravel() - operator.apply(y, mu.at_time(t=t)).to_numpy().ravel())

        self._solver = cvode.CVODE(
            cvode_rhs, 
            lmm_type='Adams', 
            nonlinsolver='fixedpoint', 
            max_steps=1000000, 
            one_step_compute=True
        )
        min_reportable_step = (end_time - initial_time)/num_values
        progress = init_progress_widget(initial_time, end_time)
        tock = tick()
        y = [self._solver.init_step(initial_time, initial_data.to_numpy().T[0]).values.y]
        t = [initial_time]
        last_t = initial_time
        while t[-1] < end_time:
            values = self._solver.step(end_time).values
            tock = update_progress_widget(progress, tock, values.t, last_t)
            last_t = values.t
            if values.t - t[-1] < min_reportable_step: continue
            t.append(values.t)
            y.append(values.y)
        t = np.array(t)
        y = np.array(y)
        progress.close()
        
        if isinstance(operator.source, XarrayVectorSpace):
            return [operator.source.from_numpy(y, extended_coord_data={'Time': t})]
        return ((operator.source.from_numpy(u), t) for u, t in zip(y, t))
            
        # self._solver = init_cvode(cvode_rhs, num_values)
        # t_list = np.linspace(initial_time, end_time, num_values)
        # sol = solve_cvode(self._solver, t_list, initial_data)
        # progress.close()
        # if isinstance(operator.source, XarrayVectorSpace):
        #     return [operator.source.from_numpy(sol.values.y, extended_coord_data={'Time': t_list})]
        # return ((operator.source.from_numpy(u), t) for u, t in zip(sol.values.y, t_list))

# %% ../../nbs/api/pymor/timestepping.ipynb
class BDFTimeStepper(TimeStepper):
    def __init__(self):
        # self.__auto_init(locals())
        pass
        
    def iterate(self, initial_time, end_time, initial_data, operator, rhs=None, mass=None, mu=None, num_values=None):
        options = cvode_solver_options()['cvode_bdf']
        # if solver_options:
        #     options.update(solver_options)
        
        def cvode_rhs(t, y, ydot):
            np.copyto(ydot, (-operator.assemble(mu.at_time(t=t)).apply(operator.source.from_numpy(y)) + rhs.as_range_array(mu.at_time(t=t))).to_numpy().T[0])

        def preconditioner_setup(t, y, jok, jcurPtr, gamma, user_data):
            """Generate P and do ILU decomposition."""
            if jok:
                jcurPtr.value = False
            else:
                user_data['approximate_jacobian'] = -to_matrix(operator.assemble(mu.at_time(t=t)))
                # if operator.solver_options is not None and 'inverse' in operator.solver_options:
                #     user_data['approximate_jacobian'] = restrict_bandwidth(
                #         user_data['approximate_jacobian'], 
                #         operator.solver_options['inverse']['preconditioner_bandwidth']
                #     )
                jcurPtr.value = True
            # Scale jacobian by -gamma, add identity matrix and do LU decomposition
            p = -gamma*user_data['approximate_jacobian'] + sparse2d_identity(user_data['approximate_jacobian'].shape[0])
            if sps.issparse(p): 
                p = p.tocsc()
            else:
                p = sps.csc_array(p)
            user_data['factored_preconditioner'] = spilu(p) # , permc_spec='NATURAL')
            return 0

        def preconditioner_solve(t, y, r, z, gamma, delta, lr, user_data):
            """ Solve the block-diagonal system Pz = r. """
            np.copyto(z, user_data['factored_preconditioner'].solve(r))
            return 0              
        
        self._solver = cvode.CVODE(
            cvode_rhs,
            lmm_type='BDF', 
            nonlinsolver='newton', 
            linsolver='spgmr',
            precond_type='left',
            prec_setupfn=preconditioner_setup, 
            prec_solvefn=preconditioner_solve,
            rtol=options['rtol'], 
            atol=options['atol'], 
            max_steps=options['max_steps'],
            user_data={},
            one_step_compute=True
        )
        # t_list = np.linspace(initial_time, end_time, num_values)
        progress = init_progress_widget(initial_time, end_time)
        tock = tick()
        # last_t = 0
        y = [self._solver.init_step(initial_time, initial_data.to_numpy().T[0]).values.y]
        t = [initial_time]
        while t[-1] < end_time:
            values = self._solver.step(end_time).values
            t.append(values.t)
            y.append(values.y)
            tock = update_progress_widget(progress, tock, t[-1], t[-2])
        t = np.array(t)
        y = np.array(y)
        # sol = self._solver.solve(t_list, initial_data.to_numpy().T[0])
        progress.close()
        
        if isinstance(operator.source, XarrayVectorSpace):
            return [operator.source.from_numpy(y, extended_coord_data={'Time': t})]
        return ((operator.source.from_numpy(u), t) for u, t in zip(y, t))

# %% ../../nbs/api/pymor/timestepping.ipynb
class StiffSwitchingTimeStepper(TimeStepper):
    def __init__(self):
        # self.__auto_init(locals())
        pass
        
    def iterate(self, initial_time, end_time, initial_data, operator, rhs=None, mass=None, mu=None, num_values=None, solver_options=None):
        options = cvode_solver_options()['cvode_stiff_switching']
        if solver_options:
            options.update(solver_options)

        if initial_time < options['t_switch']:
            nonstiff_solver = AdamsTimeStepper()
            nonstiff = list(nonstiff_solver.iterate(initial_time, options['t_switch'], initial_data, operator, rhs=rhs, mass=mass, mu=mu, num_values=num_values))
            nonstiff_t_list = nonstiff_solver._t_list
            initial_data = nonstiff[-1][0]
            initial_time = nonstiff_t_list[-1]
        else:
            nonstiff = []
            nonstiff_t_list = []

        if end_time > options['t_switch']:
            stiff_solver = BDFTimeStepper()
            stiff = list(stiff_solver.iterate(options['t_switch'], end_time, initial_data, operator, rhs=rhs, mass=mass, mu=mu, num_values=num_values, solver_options=solver_options))
            stiff_t_list = stiff_solver._t_list            
        else:
            stiff = []
            stiff_t_list = []

        self._t_list = np.concatenate([nonstiff_t_list, stiff_t_list])
        
        return nonstiff + stiff



"""Extended functionality for [pyMOR](https://pymor.org/) vector arrays"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/pymor/vectorarrays.ipynb.

# %% auto 0
__all__ = ['Lbl', 'unlabel_dims', 'default_coordinate', 'dataarray_coordinate_data', 'validate_coord_data', 'coord_data_shape',
           'make_coordinates', 'XarrayVectorSpace', 'transpose_like', 'XarrayVectorArrayImpl', 'XarrayVectorArray',
           'plotly_dataarray']

# %% ../../nbs/api/pymor/vectorarrays.ipynb
from numbers import Number, Real
from functools import partial, cached_property, singledispatch
import inspect
from enum import StrEnum
import math

from fastcore.basics import patch, first
from fastcore.meta import delegates

import numpy as np
from numpy import array, ndarray
from sparse import SparseArray
import xarray as xr
from xarray import DataArray, Variable
from pymor.basic import NumpyVectorSpace
from pymor.core.base import classinstancemethod
from pymor.vectorarrays.interface import VectorArray, VectorArrayImpl, VectorSpace

from ..utilities.basic import get_item, filter_args, filter_out_args
from ..utilities.xarray import Coordinates
from ..utilities.formatting import prefix_format

# %% ../../nbs/api/pymor/vectorarrays.ipynb
import pandas as pd
import plotly.express as px
import plotly

# %% ../../nbs/api/pymor/vectorarrays.ipynb
class Lbl(StrEnum):
    SRC = ' (source)'
    RNG = ' (range)'

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def unlabel_dims(obj):
    match obj:
        case str(): return obj.removesuffix(Lbl.SRC).removesuffix(Lbl.RNG)
        case dict(): return {unlabel_dims(k): v for k,v in obj.items()}
        case Coordinates(): return Coordinates({unlabel_dims(k): v.data for k,v in obj.items()})
        case DataArray(): return obj.rename({dim: unlabel_dims(dim) for dim in obj.dims})
        case XarrayVectorSpace(): return obj.with_(coord_data={unlabel_dims(k): v for k,v in obj.coord_data.items()})
    return obj

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def default_coordinate(dim:str='len', size:int|list[int]|tuple[int]|None=None) -> tuple:
    if isinstance(size, list|tuple): size = size[0]
    if size: return (np.arange(size, dtype=int),)
    return ([dim],)

def dataarray_coordinate_data(da:DataArray, dim:str) -> tuple:
    if dim in da.coords:
        return (da[dim].data, da[dim].attrs)
    return default_coordinate(dim, da.sizes[dim])

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@singledispatch
def validate_coord_data(obj, shape=None) -> dict:
    raise TypeError(f'Object {obj} not recognized as coordinates')

@validate_coord_data.register
def ____(obj:DataArray, shape=None):
    return {dim: dataarray_coordinate_data(obj, dim) for dim in obj.dims}

@validate_coord_data.register
def ____(obj:Coordinates, shape=None):
    return {k: (v.data, v.attrs) for k,v in obj.items()}

@validate_coord_data.register
def ____(obj:dict, shape=None):
    return {k: v if isinstance(v, tuple) else (v,) for k,v in obj.items()}

@validate_coord_data.register
def ____(obj:str, shape=None):
    return {obj: default_coordinate(obj, shape)}

@validate_coord_data.register
def ____(obj:int, shape=None): # Need int case because some internal pyMOR code calls .zeros() with integer argument
    return {obj: default_coordinate(None, obj)}

@validate_coord_data.register
def ____(obj:list, shape=None):
    if not isinstance(shape, list|tuple): raise ValueError(f'Shape must be supplied if list is used for coordinates.')
    return {dim: default_coordinate(dim, size) for dim,size in zip(obj, shape)}

@validate_coord_data.register
def ____(obj:None, shape=None):
    return {}

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def coord_data_shape(coord_data):
    return [len(c[0]) for c in coord_data.values()]

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def make_coordinates(coord_data:dict) -> Coordinates:
    return Coordinates({k:(k, *v) for k,v in coord_data.items()})

# %% ../../nbs/api/pymor/vectorarrays.ipynb
class XarrayVectorSpace(VectorSpace):
    """`VectorSpace` of `XarrayVectorArrays`."""

    def __init__(
        self, 
        coord_data:Coordinates|dict|DataArray|str|None=None,
        name:str|None=None, 
        attrs:dict|None=None, 
        id=None      # See `~pymor.vectorarrays.interface.VectorSpace.id`.
    ):
        if name is None and hasattr(coord_data, 'name'):
            name = coord_data.name
        coord_data = validate_coord_data(coord_data)
        self.__auto_init(locals())
        self.dims = list(self.coord_data)
        self.ndim = len(self.dims)
        self.shape = coord_data_shape(self.coord_data)
        self.size = math.prod(self.shape)

    @cached_property
    def _array(self):
        return DataArray(
            np.zeros(self.shape),
            dims=self.dims,
            coords=make_coordinates(self.coord_data), 
            name=self.name, 
            attrs=self.attrs
        ) 
    
    @cached_property
    def coords(self):
        return self._array.coords
    
    @property
    def dim(self): return self.size
    
    @classmethod
    def _array_factory(cls, data, space, data_dims=None, extended_coord_data=None):
        return XarrayVectorArray(space, XarrayVectorArrayImpl(data, space, data_dims, extended_coord_data))

    @classinstancemethod
    def make_array(cls, obj, id=None):
        vec = cls._array_factory(obj, id=id)
        # vec.name = obj.name
        return vec

    @make_array.instancemethod
    def make_array(self, data, data_dims=None, extended_coord_data=None):
        """:noindex:"""
        return self._array_factory(data, self, data_dims, extended_coord_data)
    
    def zeros(
        self, 
        coords:dict|None=None, # Coordinates to extend vector space over
        reserve=0
    ): 
        raise NotImplementedError # Defined below
    
    def ones(
        self, 
        coords:dict|None=None, # Coordinates to extend vector space over
        reserve=0
    ): 
        raise NotImplementedError # Defined below

    # def __str__(self):
    #     return _coords_str(self.coords)

    # def short_str(self):
    #     return _short_coords_str(self.coords)
    
    def _repr_html_(self, show_details=True):
        if not show_details:
            return ' ⛒ '.join(f'<b>{dim}</b>({length})' for dim,length in zip(self.dims, self.shape))
        return f'''
            <details closed>
                <summary>
                    {self._repr_html_(show_details=False)}
                </summary>
                {self._array._repr_html_()}
            </details>
        '''    
    
    def __hash__(self):
        return hash(self.dim)

    @property
    def is_scalar(self):
        return self.dim == 1

# %% ../../nbs/api/pymor/vectorarrays.ipynb
# @patch(as_prop=True)
# def ndim(self:XarrayVectorSpace): 
#     """Number of dimensions of the vector space."""
#     return self._array.ndim

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def rename(
    self:XarrayVectorSpace,
    name,
):
    return self.with_(name=name) #XarrayVectorSpace(self._array.rename(new_name_or_name_dict, **names))

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def __eq__(self:XarrayVectorSpace, other):
    return (
        type(other) is type(self)
        and all(k1 == k2 and np.array_equal(v1[0], v2[0]) for (k1,v1),(k2,v2) in zip(self.coord_data.items(), other.coord_data.items()))
        ## NEED TO FIX THIS
        # and all((self.coords_dict[k] == other.coords_dict[k]) for k in self.coords_dict)
    )

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def __mul__(self:XarrayVectorSpace, other):
    if not isinstance(other, XarrayVectorSpace): raise NotImplementedError
    return XarrayVectorSpace(self.coord_data | other.coord_data, name=other.name)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def transpose_like(a:ndarray, dims, new_dims):
    return a.transpose([dims.index(d) for d in new_dims])

# %% ../../nbs/api/pymor/vectorarrays.ipynb
class XarrayVectorArrayImpl(VectorArrayImpl):

    def __init__(
        self, 
        _data:ndarray|DataArray, 
        _space, 
        _data_dims=None, 
        _extended_coord_data:Coordinates|dict|None=None
    ):
        if isinstance(_data, DataArray):
            _data_dims = _data.dims
            _extended_coord_data = _data
            _data = _data.data
        if isinstance(_data, SparseArray): 
            _data = _data.todense()
        if _data_dims is None:
            _data_dims = _space.dims
        _extended_coord_data = validate_coord_data(_extended_coord_data)
        _extended_coord_data = {k:v for k,v in _extended_coord_data.items() if k not in _space.dims}
        self._dims = list(_extended_coord_data)
        all_dims = self._dims + _space.dims
        if set(all_dims) != set(_data_dims): raise ValueError(f'Data dims {_data_dims} do not match specified dims {all_dims}')
        if len(all_dims) != _data.ndim: raise ValueError(f'Dimensions {all_dims} do not match number of axes of data of shape {_data.shape}.')
        _data = transpose_like(_data, _data_dims, all_dims)
        _data_dims = all_dims
        self.__auto_init(locals())
        self._shape = _data.shape
        self._sizes = dict(zip(_data_dims, self._shape))
        self._array = None
        self._extended_shape = coord_data_shape(self._extended_coord_data)
        self._extended_size = math.prod(self._extended_shape)

    @property
    def array(self): 
        if self._array is None:
            self._array = DataArray(self._data, dims=self._data_dims, coords=make_coordinates(self._extended_coord_data | self._space.coord_data))
        return self._array

    @property
    def _coords(self): return self.array.coords

    @property
    def _len(self): return self._data.size // self._space.size
    
    def __len__(self): return self._len

    def copy(self, deep, ind):
        _data = self._data if ind is None else self._data[ind].copy()
        return XarrayVectorArrayImpl(_data, self._space, self._data_dims, self._extended_coord_data)
    
    def to_numpy(self, ensure_copy, ind): raise NotImplementedError
    def real(self, ind): raise NotImplementedError
    def imag(self, ind): raise NotImplementedError
    def conj(self, ind): raise NotImplementedError
    def delete(self, ind): raise NotImplementedError
    def append(self, other, remove_from_other, oind): raise NotImplementedError
    def scal(self, alpha, ind): raise NotImplementedError
    def scal_copy(self, alpha, ind): raise NotImplementedError
    def axpy(self, alpha, x, ind, xind): raise NotImplementedError
    def axpy_copy(self, alpha, x, ind, xind): raise NotImplementedError
    def inner(self, other, ind, oind): raise NotImplementedError
    def pairwise_inner(self, other, ind, oind): raise NotImplementedError
    def lincomb(self, coefficients, ind): raise NotImplementedError
    def norm(self, ind): raise NotImplementedError
    def norm2(self, ind): raise NotImplementedError
    def dofs(self, dof_indices, ind): raise NotImplementedError
    def amax(self, ind): raise NotImplementedError

# %% ../../nbs/api/pymor/vectorarrays.ipynb
class XarrayVectorArray(VectorArray):
    """`VectorArray` implementation via xarray arrays."""

    impl_type = XarrayVectorArrayImpl

    @property
    def data(self): return self.impl._data
    
    @property
    def array(self): return self.impl.array

    @property
    def shape(self): return self.impl._shape
    
    @property
    def coords(self): return self.impl._coords

    @property
    def dims(self): return self.impl._space.dims
    
    @property
    def extended_dims(self): return self.impl._dims

    @property
    def extended_coord_data(self): return self.impl._extended_coord_data

    @property
    def extended_shape(self): return self.impl._extended_shape
    
    # def _coords_str(self, dims):
    #     if not dims: return '{1}'
    #     sizes = self.impl._sizes
    #     return '{' + ' ⨉ '.join(f'{dim}({sizes[dim]})' for dim in dims) + '}'        
    
    # def __str__(self):
    #     return self._coords_str(self.dims)[:-1] + ", " + self._coords_str(self.extended_dims)[1:]
    
    # def short_str(self):
    #     return _short_coords_str(self.space)[:-1] + ", " + _short_coords_str(self.coords, self.dims)[1:]
    
    def _repr_html_(self, show_details=True):
        space = self.space._repr_html_(show_details=False)
        extended = [f'{k}({len(v[0])})' for k,v in self.extended_coord_data.items()]
        if not extended: extended = ['∅']
        out = ' ⛒ '.join(extended + [space])
        if not show_details: return out
        return f'''
            <details closed>
                <summary>
                    {out}
                </summary>
                {self.array._repr_html_()}
            </details>
        '''

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def from_data(self:XarrayVectorSpace, data, data_dims=None, extended_coord_data=None):
    return self.make_array(data, data_dims, extended_coord_data)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def from_numpy(
    self:XarrayVectorSpace, 
    data:ndarray, 
    extended_coord_data:dict|str|None=None, 
    id=None, 
    ensure_copy=False
)->XarrayVectorArray: # A vector array with data from the numpy array
    """Return an `XarrayVectorArray` in the vector space with data from the `ndarray` `data`."""
    if ensure_copy: data = data.copy()
    if extended_coord_data is None:
        extended_coord_data = {}
    elif isinstance(extended_coord_data, str):
        extended_coord_data = {extended_coord_data: np.arange(data.size // self.dim)}
    extended_coord_data = validate_coord_data(extended_coord_data)
    extended_shape = coord_data_shape(extended_coord_data)
    shape = extended_shape + self.shape
    return self.from_data(data.reshape(shape), list(extended_coord_data) + self.dims, extended_coord_data)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def from_xarray(
    self:XarrayVectorSpace, 
    data:DataArray
)->XarrayVectorArray:
    """Return an `XarrayVectorArray` containing data from `DataArray` `data`."""
    # extended_coord_data = {k:v for k,v in data.coords.items() if k not in self.dims}
    return self.from_data(data)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def concatenate(self:XarrayVectorSpace, arrays, dim):
    return self.from_xarray(xr.concat([array.array for array in arrays], Variable(first(dim), first(dim.values()))))

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def array_from_fn(
    self:XarrayVectorSpace,
    fn,
    extended_coords:dict|str|None=None, # Coordinates to extend vector space over
    reserve=0,
)->XarrayVectorArray: # `XarrayVectorArray` with all elements equal to zero
    """Return `XarrayVectorArray` of null vectors in XarrayVectorSpace optionally extended to include supplied coordinates `coords`."""
    extended_coords = validate_coord_data(extended_coords)
    shape = coord_data_shape(extended_coords)
    data = fn(shape + self.shape)
    return self.from_data(data, list(extended_coords) + self.dims, extended_coords)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def zeros(
    self:XarrayVectorSpace, 
    extended_coords:dict|None=None, # Coordinates to extend vector space over
    reserve:int=0
)->XarrayVectorArray: # `XarrayVectorArray` with all elements equal to zero
    """Return `XarrayVectorArray` of null vectors in XarrayVectorSpace optionally extended to include supplied coordinates `coords`."""
    return self.array_from_fn(np.zeros, extended_coords, reserve)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def ones(
    self:XarrayVectorSpace, 
    extended_coords:dict|None=None, # Coordinates to extend vector space over
    reserve:int=0
)->XarrayVectorArray: # `XarrayVectorArray` with all elements equal to one
    """Return `XarrayVectorArray` of vectors with each element equal to one in XarrayVectorSpace optionally extended to include supplied coordinates `coords`."""
    return self.array_from_fn(np.ones, extended_coords, reserve)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
from pymor.vectorarrays.interface import _create_random_values

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def random(
    self:XarrayVectorSpace, 
    extended_coords=None, 
    distribution='uniform',
    name=None
)->XarrayVectorArray: # A random `XarrayVectorArray` in the vector space
    out = self.array_from_fn(lambda shape: _create_random_values(shape, distribution), extended_coords)
    if name is not None: out.name = name
    return out

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def __contains__(self:XarrayVectorSpace, other):
    # This is reversed because when pymor calls `assert U in op.source`
    # we really want to assert that the source space of `op` is contained in the space of `U`, 
    # which more logically would be written "assert op.source in U"
    if isinstance(other, XarrayVectorSpace):
        theirs = other.coord_data
    elif isinstance(other, XarrayVectorArray):
        theirs = other.space.coord_data
    else:
        return False
    mine = self.coord_data
    if not all(k in theirs for k in mine): return False
    if any(not np.array_equal(mine[k][0], theirs[k][0]) for k in mine): return False
    return True
    
    # if isinstance(other, XarrayVectorSpace):
    #     return other._array.coords.contain(self._array.coords)
    # elif isinstance(other, XarrayVectorArray):
    #     return other.space._array.coords.contain(self._array.coords)
    # return False

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def array_labeled_as_source(self:XarrayVectorArray):
    return self.array.rename({d: d + Lbl.SRC for d in self.dims})

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def item(self:XarrayVectorArray):
    """Return the single value stored in the array if there is only one value."""
    return self.data.item()

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch(as_prop=True)
def stacked_array(self:XarrayVectorArrayImpl):
    core = {'core': self._space.dims} if self._space.dims else {}
    extended = {'extended': self._dims} if self._dims else {}
    return self.array.stack(core | extended, create_index=False)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def stacked_data(self:XarrayVectorArrayImpl):
    return transpose_like(
        self._data, self._data_dims, self._space.dims + list(self._extended_coord_data)
    ).reshape(
        self._space.size, self._extended_size
    )

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def to_numpy(self:XarrayVectorArrayImpl, ensure_copy, ind):
    out = self.stacked_data()
    if ind is not None: out = out[:, ind]
    if ensure_copy and not out.flags['OWNDATA']:
        return out.copy()
    else:
        return out

    # array = self.stacked_array.to_numpy()
    # if not self._space.dims: array = array[None, :]
    # if not self._dims: array = array[:, None]
    # if ensure_copy and not array.flags['OWNDATA']:
    #     return array.copy()
    # else:
    #     return array

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def scal(self:XarrayVectorArrayImpl, alpha, ind):
    if ind is not None: raise NotImplementedError
    self._data *= alpha

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def scal_copy(self:XarrayVectorArrayImpl, alpha, ind):
    if ind is not None: raise NotImplementedError
    return type(self)(self._data * alpha, self._space, self._data_dims, self._extended_coord_data)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def axpy(self:XarrayVectorArrayImpl, alpha, x, ind, xind):
    if ind is not None: raise NotImplementedError
    B = x._data
    if isinstance(alpha, Number):
        if alpha == 1:
            self._data += B
            return
        elif alpha == -1:
            self._data -= B
            return

    self._data += B * alpha

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def axpy_copy(self:XarrayVectorArrayImpl, alpha, x, ind, xind):
    if ind is not None or xind is not None: raise NotImplementedError
    B = x._data
    # if issparse(B): B = B.todense()
    if isinstance(alpha, Number):
        if alpha == 1:
            return type(self)(self._data + B, self._space, self._data_dims, self._extended_coord_data)
        elif alpha == -1:
            return type(self)(self._data - B, self._space, self._data_dims, self._extended_coord_data)
    return type(self)(self._data + B * alpha, self._space, self._data_dims, self._extended_coord_data)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def __mul__(self:XarrayVectorArray, other):
    match other:
        case DataArray():
            other = unlabel_dims(other)
            if self.coords.contain(other.coords):
                out = self.space.from_xarray(self.array * other)
                out.name = self.name
                return out
            else:
                raise ValueError("DataArray coordinates are not a subset of VectorArray coordinates.")
        case XarrayVectorArray(): 
            out = (self.space * other.space).from_xarray(self.array * other.array)
            out.name = self.name
            return out
        case Number():
            return type(self)(self.space, self.impl.scal_copy(other, self.ind))
    raise NotImplementedError

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def _reim(da:DataArray):
    return xr.concat([da.real, da.imag], dim=xr.Variable('Part', ['Re', 'Im'])).transpose(*da.dims, 'Part')

# %% ../../nbs/api/pymor/vectorarrays.ipynb
@patch
def real(self:XarrayVectorArrayImpl, ind):
    return XarrayVectorArrayImpl(self.copy(False, ind)._data.real, self._space, self._data_dims, self._extended_coord_data)
    
@patch
def imag(self:XarrayVectorArrayImpl, ind):
    return XarrayVectorArrayImpl(self.copy(False, ind)._data.imag, self._space, self._data_dims, self._extended_coord_data)
    
@patch
def conj(self:XarrayVectorArrayImpl, ind):
    if np.isrealobj(self._data):
        return self.copy(False, ind)
    return XarrayVectorArrayImpl(np.conj(self.copy(False, ind)._data), self._space, self._data_dims, self._extended_coord_data)
    
@patch
def reim(self:XarrayVectorArrayImpl):
    return XarrayVectorArrayImpl(
        np.array([self._data.real, self._data.imag]), 
        self._space, ['Part'] + self._data_dims, 
        {'Part': (['Re', 'Im'], {})} | self._extended_coord_data
    )
    
@patch(as_prop=True)
def reim(self:XarrayVectorArray):
    impl = self.impl.reim()
    if impl is self.impl:
        return self.copy()
    else:
        return type(self)(self.space, impl)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
plotly.io.templates.default = "plotly_white"
plotly.io.templates['plotly_white'].layout.legend = plotly.graph_objects.layout.Legend(tracegroupgap=0)
plotly.io.templates["plotly_white"].layout.width = 700
plotly.io.templates["plotly_white"].layout.height = 400
plotly.io.templates["plotly_white"].layout.autosize = False
plotly.io.templates["plotly_white"].layout.colorway = array([
       '#2fa1da', '#fb4f2f', '#e4ae38', '#6d904f', '#8a8a8a', '#16bdcf',
       '#9367bc', '#d62628', '#1f77b3', '#e277c1', '#8c564b', '#bcbc21',
       '#3a0182', '#004200', '#0fffa8', '#5d003f', '#c6bcff', '#424f52',
       '#b80080', '#ffb6b3', '#7c0100', '#6026ff', '#ffff9a', '#aec8aa',
       '#00857c', '#543a00', '#93fbff', '#00bf00', '#7c00a0', '#aa7200',
       '#90ff00', '#01bd89', '#00447b', '#c8826e', '#ff1f82', '#dd00ff',
       '#057400', '#644460', '#878eff', '#ffb5f4', '#526236', '#cd85ff',
       '#676983', '#bdb3bd', '#a56089', '#95d3ff', '#0100f7', '#ff8001',
       '#8a2844', '#aca06d', '#52448a', '#c8ffd8', '#aa4600', '#ff798e',
       '#82d370', '#909ebf', '#9300f4', '#ebcf9a', '#ac8ab1', '#006249',
       '#ffdb00', '#877750', '#7eaaa3', '#000097', '#f400c6', '#643328',
       '#006677', '#03e2c8', '#a736ae', '#c4dbe1', '#4d6eff', '#9a9301',
       '#cd576b', '#efddfd', '#795900', '#5e879a', '#b3ff91', '#5d726b',
       '#520066', '#058750', '#831f6e', '#3b9505', '#647200', '#f0a06b',
       '#5e4f44', '#bc0049', '#cf6726', '#d695aa', '#895dff', '#826b75',
       '#2a54b8', '#6e7cba', '#e6d4d3', '#5d0018', '#7c3b01', '#80b17c',
       '#c8d87c', '#00e83b', '#7cb1ff', '#ff54ff', '#a32621', '#1ce4ff',
       '#7caf3b', '#7b4b90', '#dfff48', '#6b00c3', '#cda897', '#bd62c3',
       '#89cdcd', '#4603c8', '#5d9179', '#414901', '#05a79c', '#cf8c36',
       '#fff7cf', '#425470', '#b544ff', '#cf4993', '#cfa3df', '#93d400',
       '#a793da', '#2da557', '#8ce2b5', '#a3a89c', '#6b5bb6', '#ff7e5d',
       '#a78289', '#afbdd8', '#2ac3ff', '#a5673d', '#f690fd', '#874b64',
       '#ff0c4b', '#215d23', '#4291ff', '#87829c', '#672d44', '#b14f41',
       '#004d52', '#5e1a00', '#ac4167', '#4f3167', '#d6fffd', '#7eb5d1',
       '#a8b869', '#ff95ca', '#c87495', '#364f38', '#ffcf62', '#5d5762',
       '#879375', '#a877ff', '#03c862', '#e6bdd4', '#d4e2cf', '#876790',
       '#897c26', '#cddbff', '#aa676b', '#313474', '#ff5da8', '#009aaf',
       '#70ffdd', '#775b38', '#4f649a', '#cc00b3', '#567b54', '#506e7b',
       '#015e91', '#aabcbd', '#017e99', '#03dd97', '#873a2b', '#ef958e',
       '#75c6aa', '#70695d', '#ccdb08', '#af8556', '#d80075', '#9c3f80',
       '#d84400', '#dd6754', '#5eff79', '#d4b172', '#62265d', '#baa13d',
       '#d8f2b3', '#56018e', '#a19aaa', '#4d4926', '#a3a8ff', '#ace8db',
       '#995901', '#ac00e2', '#46822f', '#cac3ac', '#00c4b5', '#605277',
       '#336d67', '#a59180', '#8399a1', '#fd5664', '#7095d1', '#728c07',
       '#7e004b', '#152fa0', '#d1c1e2', '#c885cf', '#6b444b', '#7e0023',
       '#00a179', '#b1a8cf', '#f90000', '#afe8ff', '#939e4f', '#727982',
       '#d82d54', '#466001', '#0059ff', '#773fb5', '#ace460', '#674424',
       '#525d50', '#957267', '#a8e49a', '#a30057', '#d862f6', '#8e7ccf',
       '#ffbc93', '#a30091', '#9affb8', '#a7c1ff', '#f46200', '#e4efff',
       '#b89ca3', '#609593', '#ff9e34', '#8c2800', '#726b31', '#df824d',
       '#af7bd4', '#bc2d00', '#7b6ea3', '#484262', '#c6a3ff', '#004d28',
       '#c3c68e', '#df48d6', '#e6e864', '#e4c10a', '#00f4f0', '#9e5ba1',
       '#4b41b6', '#64338e', '#757e6b', '#a88936'])

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def _categorical_dims(array:DataArray): return [dim for dim in array.dims if not isinstance(array[dim][0].item(), Number)]
def _numerical_dims(array:DataArray):   return [dim for dim in array.dims if     isinstance(array[dim][0].item(), Number)]

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def _numerical_dim_sort_order(da, dim):
    """Assign priority for each dimension to be put on the x axis.
    Lower number is higher priority. Time is preferred on a scrubber, velocity preferred on the x-axis.
    """
    if "time" in dim.lower(): return 2
    if "velocity" in dim.lower(): return 0
    if isinstance(da[dim][0].item(), Real): return 1
    return 0

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def _set_plotly_frame(fig, n):
    fig.update_layout(sliders=[dict(active=n)])
    for i in range(len(fig.data)):
        fig.data[i].y = fig.frames[n].data[i].y

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def _format_slider_labels(fig, slider_label_precision=2):
    for step in fig.layout.sliders[0].steps:
        step.label = prefix_format(float(step.label), slider_label_precision)

# %% ../../nbs/api/pymor/vectorarrays.ipynb
def plotly_dataarray(da:DataArray, slider_label_precision=2, **kwargs):
    """Visualize the data contained in the `XarrayVectorArray`.
    Put the first dimension with numerical coordinates on the x axis, and include a scrubber if there is a second numerical dimension. 
    Put categorical dimensions in the legend. If the data are complex, add the real and imaginary parts as a categorical dimension.
    If there are no numerical dimensions, plot the categorical dimension(s) as a horizontal bar plot.
    """
    da = da.copy()
    if np.any(np.iscomplex(da)): da = _reim(da)
    if da.name is None: da = da.rename('Value')
    df = da.to_dataframe().reset_index()
    num = _numerical_dims(da)
    cat = _categorical_dims(da)
    if not num and da.ndim <= 2: 
        return px.bar(
            df, 
            y=cat[0], 
            x=da.name, 
            color=get_item(cat, 1), 
            orientation='h', 
            **filter_args(px.bar, **kwargs)
        ).update_layout(
            legend_title=None
        )
    if da.ndim > 4: raise NotImplementedError
    if len(num) > 2: num, cat = num[:2], cat + num[2:]
    if len(cat) > 2: cat, num = cat[:2], num + cat[2:]
    num = sorted(num, key=partial(_numerical_dim_sort_order, da))
    fig = px.line(
        df,
        x=get_item(num, 0),
        y=da.name,
        animation_frame=get_item(num, 1),
        color=get_item(cat, 0),
        line_dash=get_item(cat, 1),
        **filter_args(px.line, **kwargs)
    ).update_layout(
        legend_title=None
    )
    if len(num) == 2:
        padding = .05
        ymin, ymax = da.min(), da.max()
        yrange = (ymax - ymin)
        fig.update_layout(
            yaxis_range=(ymin - padding * yrange, ymax + padding * yrange)
        )
        _format_slider_labels(fig, slider_label_precision=slider_label_precision)
        _set_plotly_frame(fig, len(da[num[1]])//3)
        # Set total animation time to 3s, unless that would make individual frames shorter than 100ms, which seems to be performance limit
        fig.layout.updatemenus[0].buttons[0].args[1]['frame']['duration'] = max(3000/len(da[num[1]]), 100) 
        fig.layout.updatemenus[0].buttons[0].args[1]['transition']['duration'] = 0
    kwargs = filter_out_args(px.line, **kwargs)
    for k in kwargs:
        try: fig.update_layout({k: kwargs[k]})
        except ValueError: pass
        try: fig.update_traces({k: kwargs[k]})
        except ValueError: pass
    return fig

# %% ../../nbs/api/pymor/vectorarrays.ipynb
## Could put plotting hints in the attrs of the coordinates inside the model
@patch
@delegates(plotly_dataarray)
def visualize(self:XarrayVectorArray, **kwargs):
    """Visualize the data contained in the `XarrayVectorArray`.
    Put the first dimension with numerical coordinates on the x axis, and include a scrubber if there is a second numerical dimension.
    Atomic velocity is prioritized to be on the x axis, time is prioritized to be on the scrubber.
    Put categorical dimensions in the legend. If the data are complex, add the real and imaginary parts as a categorical dimension.
    If there are no numerical dimensions, plot the categorical dimension(s) as a horizontal bar plot.
    """
    return plotly_dataarray(self.array, **kwargs)



"""Testing utilities"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/utilities/testing.ipynb.

# %% auto 0
__all__ = ['test_array', 'test_da']

# %% ../../nbs/api/utilities/testing.ipynb
import numpy as np
import xarray as xr
from fastcore.test import test_close

# %% ../../nbs/api/utilities/testing.ipynb
def test_array(file, name, arr, update='allow'):
    file = file + '_test.npz'
    try: arrays = dict(np.load(file))
    except FileNotFoundError: 
        if update != 'forbid': arrays = {}
        else: raise
    if name in arrays and update != 'force': test_close(arrays[name], arr)
    elif update != 'forbid':
        print(f'adding {name} array')
        np.savez(file, **(arrays | {name: arr}))
    else: raise RuntimeError(f'Test array {name} not in .npz file {file}.')

# %% ../../nbs/api/utilities/testing.ipynb
def test_da(file, name, da, update='forbid'):
    file = file + '_test.cdf'
    if update != 'force':
        try:
            with xr.open_dataarray(file, group=name) as saved: 
                assert saved.equals(da)
                return
        except AssertionError:
            raise AssertionError(f'Saved DataArray\n{saved}\nnot equal to supplied DataArray\n{da}.')
        except FileNotFoundError: 
            if update != 'allow': raise
        except OSError:
            if update != 'allow': raise RuntimeError(f'Test array {name} not in file {file}.')
    print(f'adding {name} DataArray to {file}')
    da.to_netcdf(file, group=name, mode='a')



"""Extra functionality for [NumPy](https://numpy.org/)"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/utilities/numpy.ipynb.

# %% auto 0
__all__ = ['sym_range']

# %% ../../nbs/api/utilities/numpy.ipynb
import numpy as np
from numpy import array, ndarray

# %% ../../nbs/api/utilities/numpy.ipynb
def sym_range(
    n # Half-length of the array
)->ndarray: # An integer NumPy array running from -n to n
    """A symmetric `arange`."""
    return np.arange(-n, n+1)



"""Define and solve models for LGS systems"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/lgssystem.ipynb.

# %% auto 0
__all__ = ['optree', 'ScipyGCROTMKSpILUSolver', 'LGSSystem']

# %% ../nbs/api/lgssystem.ipynb
import importlib
from IPython.display import SVG, Markdown

from pymor.vectorarrays.interface import VectorArray
from pymor.operators.interface import Operator
from pymor.basic import IdentityOperator, VectorOperator

from .imports import *
from .utilities.nbdev import DictTbl, AttributeTbl
from .utilities.testing import test_array
from .utilities.sparse import sparse_kronecker_matrix, sparse_toeplitz, sparse_identity, sparse_diag, sparse
from .utilities.numpy import sym_range
from .pymor.parameters import *
from .pymor.vectorarrays import *
from .pymor.operators import *
from .pymor.timestepping import *
from .pymor.grids import *
from .pymor.models import *
from .velocitygroups import *
from .dmelements import *

# %% ../nbs/api/lgssystem.ipynb
def optree(op):
    match op:
        case LincombOperator(): 
            return [op.name + "(Lincomb)", [optree(o) for o in op.operators]]
        case ProductOperator(): 
            return [op.name + "(Product)", [optree(o) for o in op.operators]]
        # case IdentityOperator():
        #     return "Identity"
        case _:
            return f'{op.name}({op.__class__.__name__.removesuffix("Operator")})'

# %% ../nbs/api/lgssystem.ipynb
from scipy.sparse.linalg import LinearOperator, gcrotmk, spilu
from pymor.solvers.interface import Solver

# %% ../nbs/api/lgssystem.ipynb
class ScipyGCROTMKSpILUSolver(Solver):

    # @defaults('tol', 'maxiter', 'spilu_drop_tol', 'spilu_fill_factor', 'spilu_drop_rule', 'spilu_permc_spec')
    # @defaults('check_finite')
    def __init__(self, check_finite=None, atol=0., rtol=1e-5, maxiter=1000,
                 spilu_drop_tol=1e-4, spilu_fill_factor=20, spilu_drop_rule=None, spilu_permc_spec='COLAMD', preconditioner=None):
        super().__init__()
        self.__auto_init(locals())

    def _solve(self, operator, V, mu, initial_guess):
        operator = operator.assemble(mu)
        matrix = to_matrix(operator).tocsc()
        precond = to_matrix(self.preconditioner.assemble(mu)).tocsc() if self.preconditioner else matrix
            
        V = V.to_numpy()
        initial_guess = initial_guess.to_numpy() if initial_guess is not None else None
        promoted_type = np.promote_types(matrix.dtype, V.dtype)

        R = np.empty((matrix.shape[1], V.shape[1]), dtype=promoted_type, order='F')
        ilu = spilu(precond, drop_tol=self.spilu_drop_tol, fill_factor=self.spilu_fill_factor,
                    drop_rule=self.spilu_drop_rule, permc_spec=self.spilu_permc_spec)
        precond = LinearOperator(matrix.shape, ilu.solve)
        for i in range(V.shape[1]):
            R[:, i], info = gcrotmk(matrix, V[:, i], initial_guess[:, i] if initial_guess is not None else None,
                                     atol=self.atol, rtol=self.rtol, maxiter=self.maxiter, M=precond)
            if info != 0:
                if info > 0:
                    raise InversionError(f'gcrotmk failed to converge after {info} iterations')
                else:
                    raise InversionError(f'gcrotmk failed with error code {info} (illegal input or breakdown)')

        if self.check_finite:
            if not np.isfinite(np.sum(R)):
                raise InversionError('Result contains non-finite values')

        return operator.source.from_numpy(R), {}

# %% ../nbs/api/lgssystem.ipynb
class LGSSystem(ParametricObject):
    """`LGSSystem` creates numerical models for the chosen laser guide star atomic system."""
    
    block_names = [
        "A_ind", # Part of density-matrix evolution matrix that is independent of atomic velocity
        "A_dop", # Part of density-matrix evolution matrix describing Doppler shift
        "A_rec", # Part of density-matrix evolution matrix describing atomic recoil
        "A_vcc", # Part of density-matrix evolution matrix describing velocity-changing collisions
        "b", # Constant vector in the density-matrix evolution equation describing atomic transit
        "Flux" # Operator giving return flux from each transition
    ]
    
    def __init__(
        self, 
        system:str, # Name of the atomic system to be loaded from file
        fixed_params:dict={} # Values for system parameters that will be held fixed
    ):
        self.path = LGSSystem._get_path(system)
        fixed_params = Mu(fixed_params)
        self.__auto_init(locals())
        for name in self.block_names:
            setattr(
                self, 
                name,
                LincombOperator.from_file(self.path/name).partial_assemble(self.fixed_params).with_(name=name)
            )
        self.Flux = XarrayMatrixOperator(self.Flux.matrix.rename({"Transition": "Transition (range)"}))
        self.parameters = Parameters.of(*(getattr(self, name) for name in self.block_names))
        self.initial = LincombOperator.from_file(self.path/'b').partial_assemble(Mu(BeamTransitRatePerS=1)).as_range_array()
        # self.initial = self.Flux.source.from_xarray(
        #     LincombOperator.from_file(self.path/'b').partial_assemble(Mu(BeamTransitRatePerS=1)).matrix.rename({'Density matrix (range)': 'Density matrix (source)'})
        # )
        self.n_variables = self.Flux.source.dim 
        self.dm_elements = self.Flux.assemble().matrix['Density matrix' + Lbl.SRC].rename({"Density matrix (source)": "Density matrix"})
        self.level_population = level_population(self.dm_elements)
        self.population = population(self.dm_elements) 
        self.total_population = total_population(self.dm_elements)

# %% ../nbs/api/lgssystem.ipynb
@patch(cls_method=True)
def _is_valid_dir(cls:LGSSystem, dir):
    """Check if the supplied directory has the required files to define an LGSSystem."""
    dir = Path(dir)
    if not dir.is_dir(): return False
    files = {f.name for f in dir.glob("*.mtxn")}
    required_files = {'A_dop.mtxn', 'A_ind.mtxn', 'A_rec.mtxn', 'A_vcc.mtxn', 'Flux.mtxn', 'b.mtxn'}
    return required_files <= files

# %% ../nbs/api/lgssystem.ipynb
@patch(cls_method=True)
def _builtin_paths(cls:LGSSystem):
    return [dir for dir in importlib.resources.files("pylgs.systems").iterdir() if LGSSystem._is_valid_dir(dir)]

# %% ../nbs/api/lgssystem.ipynb
@patch(cls_method=True)
def builtins(cls:LGSSystem):
    return [dir.name for dir in LGSSystem._builtin_paths()]

# %% ../nbs/api/lgssystem.ipynb
@patch(cls_method=True)
def _get_path(cls:LGSSystem, name):
    if LGSSystem._is_valid_dir(name): return Path(name)
    for dir in LGSSystem._builtin_paths(): 
        if dir.name == name: return dir
    raise FileNotFoundError(f"{name} does not correspond to a built-in or supplied LGSSystem.")

# %% ../nbs/api/lgssystem.ipynb
@patch(cls_method=True)
def diagram(
    cls:LGSSystem, 
    name:str | Path, # Name of built-in LGS system definition, or directory of supplied system
    kind="Toy" # "Toy" to show level diagram and pumped transitions, "ToScale" to show hyperfine structure to scale, "NotToScale" to show unscaled hf structure
):
    """Draw a level diagram for the system of the specified kind."""
    return SVG(LGSSystem._get_path(name)/f"LevelDiagram{kind}.svg")

# %% ../nbs/api/lgssystem.ipynb
@patch(cls_method=True)
def info(
    cls:LGSSystem, 
    name:str | Path, # Name of built-in LGS system definition, or directory of supplied system
):
    """Display information about LGS system `name`."""
    return Markdown(filename=LGSSystem._get_path(name)/"info.md")

# %% ../nbs/api/lgssystem.ipynb
@patch
def _reaction(self:LGSSystem, vg:VelocityGroups):
    return (
        (vg.identity().with_(name="VG") * self.A_ind).with_(name="VG ind.") 
        + (vg.velocity_diagonal().with_(name="Velocity") * self.A_dop).with_(name="VG dop.")
    ).with_(name="Reaction")

# %% ../nbs/api/lgssystem.ipynb
@patch
def _vcc(self:LGSSystem, vg:VelocityGroups):
    return vg.n_times_1().with_(name="MaxBoltz") * self.A_vcc

# %% ../nbs/api/lgssystem.ipynb
@patch
def _recoil(self:LGSSystem, vg:VelocityGroups):
    return vg.drho_dv().with_(name='drho_dv') * self.A_rec

# %% ../nbs/api/lgssystem.ipynb
@patch
def operator(
    self:LGSSystem, 
    vg:VelocityGroups # `VelocityGroups` specification
)->Operator:
    """The evolution operatior for the system, given the `VelocityGroups` `vg`."""
    return (-1 * (self._reaction(vg) + self._vcc(vg) + self._recoil(vg))).with_(
        name='DM evolution',
        # solver_options={'inverse': {'type': 'scipy_lgmres_spilu', 'preconditioner_bandwidth': self.n_variables}}
    )

# %% ../nbs/api/lgssystem.ipynb
@patch
def rhs(
    self:LGSSystem, 
    vg:VelocityGroups # `VelocityGroups` specification
)->VectorArray:
    """The right-hand side vector for this system given the `VelocityGroups` `vg`."""
    return velocity_density_vector(vg) * self.b.as_range_array()

# %% ../nbs/api/lgssystem.ipynb
@patch
def _initial_dm(self:LGSSystem, vg):
    return velocity_density_vector(vg) * self.initial

# %% ../nbs/api/lgssystem.ipynb
@patch
def _products(self:LGSSystem, vg): 
    return {
        'population': self.total_population, # Fraction of total atomic population in each velocity group
        'total_population': vg.sum() * self.total_population, # Total atomic population (should equal 1)
        'population_distribution': vg.normalize() * self.total_population, # Atomic population per unit velocity interval
        'level_population': self.level_population, # Population of each atomic level in each velocity group
        'total_level_population': vg.sum() * self.level_population, # Total atomic population in each level
        'level_population_distribution': vg.normalize() * self.level_population, # Level population per unit velocity interval
        'flux': self.Flux, # Return flux on each transition from each velocity group
        'total_flux': vg.sum() * self.Flux, # Total return flux
        'flux_distribution': vg.normalize() * self.Flux, # Return flux per unit velocity interval
        'velocity_sum': vg.sum(), # Sum over velocity groups
        'velocity_normalize': vg.normalize(), # Normalize by velocity-group width
    }

# %% ../nbs/api/lgssystem.ipynb
@patch
def stationary_model(
    self:LGSSystem, 
    vg:int|Iterable|VelocityGroups=6, # Specification of velocity groups 
    # output='total_flux' # Operator to be applied to solution to obtain the output
) -> StationaryModel: # Model for the steady-state LGS system
    """Return a pyMOR `StationaryModel` for the steady state cw LGS system."""
    vg = VelocityGroups(vg)
    return StationaryModel(
        self.operator(vg).with_(solver=ScipyGCROTMKSpILUSolver(spilu_fill_factor=40, preconditioner=-self._reaction(vg))),
        self.rhs(vg), 
        products=self._products(vg),
        data={
            'velocity_groups': vg,
            'grid': OnedUnstructuredGrid(vg.edges)
        }
    )

# %% ../nbs/api/lgssystem.ipynb
@patch
def adaptive_stationary_model(
    self:LGSSystem, 
    mu, # Parameter values to use during velocity-group refinement
    vg=6, # Initial velocity groups
    max_weight=0.01 # Maximum fraction of the return flux from any one velocity group
)->StationaryModel: # Steady-state LGS model with more velocity groups in regions that produce more return flux
    """Create a `StationaryModel` with narrower velocity bins in velocity regions with higher flux."""
    vg = VelocityGroups(vg)
    while True:
        m = self.stationary_model(vg)
        u = m.solve(mu)
        o = m.flux(u).array.sum('Transition').squeeze().data
        o /= np.sum(o)
        mask = o > max_weight
        if not np.sum(mask):
            break
        vg = vg.subdivide(mask)
    return self.stationary_model(vg)

# %% ../nbs/api/lgssystem.ipynb
@patch
def instationary_model(
    self:LGSSystem, 
    vg:int|Iterable|VelocityGroups=6, # Specification of velocity groups 
    # output:str='total_flux',  # Operator to be applied to solution to obtain the output
    T=1.e-6, # Final time (seconds)
    num_values=100, # Number of time step values to return
    initial_rho=None, # Initial density matrix. `None` means to use the thermal distribution
    time_stepper=BDFTimeStepper() # Solver to use
) -> InstationaryModel: # Model for the LGS system dynamics
    """Create an `InstationaryModel` for the LGS system dynamics."""
    vg = VelocityGroups(vg)
    return InstationaryModel(
        T,
        ifnone(initial_rho, self._initial_dm(vg)),
        self.operator(vg), 
        self.rhs(vg), 
        mass=None,
        time_stepper=time_stepper,
        num_values=num_values,
        # output_functional=self._products(vg)[output],
        products=self._products(vg),
        data={
            'velocity_groups': vg,
            'grid': OnedUnstructuredGrid(vg.edges)
        }
    )

# %% ../nbs/api/lgssystem.ipynb
def _to_nmo(op, format='dense'):
    return NumpyMatrixOperator(to_matrix(op, format=format))

# %% ../nbs/api/lgssystem.ipynb
@patch
@delegates(LGSSystem.instationary_model)
def numpy_instationary_model(self:LGSSystem, vg, **kwargs):
    model = self.instationary_model(vg, **kwargs)
    operator = expand(model.operator)
    operator = contract(LincombOperator([_to_nmo(o) for o in operator.operators], operator.coefficients))
    rhs = _to_nmo(model.rhs)
    initial_data = _to_nmo(model.initial_data)
    vg = model.data['velocity_groups']
    # vg_sum = XarrayMatrixOperator(np.ones([1, vg['VGNumber']]), source=vg.sum().source, range=vg.sum().range)
    output_functional = _to_nmo(IdentityOperator(vg.sum().source) * model.products['flux'])
    return InstationaryModel(
        T=model.T,
        initial_data=initial_data,
        operator=operator,
        rhs=rhs,
        output_functional=output_functional,
        mass=None,
        time_stepper=model.time_stepper,
        num_values=model.num_values,
        products=None,
        error_estimator=None,
        visualizer=None,
        name=None,
    )

# %% ../nbs/api/lgssystem.ipynb
@patch
def _floquet_products(self:LGSSystem, vg, n_vector, T): 
    return {
        # 'reconstruct': floquet_reconstruction_operator(n_vector, T),
        # 'total_population': velocity_sum(vg) * self.total_population, 
        # 'population_distribution': velocity_normalize(vg) * self.total_population, 
        'reconstructed_level_population': floquet_reconstruction_operator(n_vector, T) * vg.sum() * self.level_population, 
        # 'level_population_distribution': velocity_normalize(vg) * self.level_population, 
        # 'flux': self.flux,
        'flux_distribution': vg.normalize() * self.Flux,
        'reconstructed_flux_distribution': floquet_reconstruction_operator(n_vector, T) * vg.normalize() * self.Flux,
        'reconstructed_total_flux': floquet_reconstruction_operator(n_vector, T) * (vg.sum() * self.Flux),
        # 'velocity_sum': velocity_sum(vg),
        # 'velocity_normalize': velocity_normalize(vg),
    }

# %% ../nbs/api/lgssystem.ipynb
@patch
def stationary_floquet_model(
    self:LGSSystem, 
    mu, # Parameter specification for parameters modulated at a frequency `nu` 
    time, # List of time values to use for reconstruction
    n_vector, # Number of terms in the system Fourier series
    nu:str='nu', # Name of the frequency variable
    n_operator=None, # Number of terms in the operator Fourier series. `None` defaults to `n_vector` - 1
    vg:int|Iterable|VelocityGroups=6, # Specification of velocity groups
    solver=None,
) -> StationaryFloquetModel: # Model for the Fourier coefficients of the periodic LGS system
    """Return a `StationaryFloquetModel` for the Fourier coefficients of the periodic state of a modulated LGS system."""
    vg = VelocityGroups(vg)
    return StationaryFloquetModel(
        operator=expand(self.operator(vg)),
        rhs=self.rhs(vg), 
        mu=mu,
        nu=nu,
        time=time,
        n_vector=n_vector,
        n_operator=n_operator,
        products=self._products(vg) | self._floquet_products(vg, n_vector, time),
        data={
            'velocity_groups': vg,
            'grid': OnedUnstructuredGrid(vg.edges)
        },
        solver=solver or ScipyGCROTMKSpILUSolver(spilu_fill_factor=40),
        preconditioner=-self._reaction(vg)
    )

# %% ../nbs/api/lgssystem.ipynb
@patch
def adaptive_stationary_floquet_model(
    self:LGSSystem, 
    mu, # Parameter specification for parameters modulated at a frequency `nu`
    mu_refine,
    time, # List of time values to use for reconstruction
    n_vector, # Number of terms in the system Fourier series
    nu:str='nu', # Name of the frequency variable
    n_operator=None, # Number of terms in the operator Fourier series. `None` defaults to `n_vector` - 1
    vg=6, # Initial velocity groups
    solver=None,
    max_weight=0.01, # Maximum fraction of the return flux from any one velocity group
    subdivisions=4,
)->StationaryFloquetModel: # Steady-state LGS model with more velocity groups in regions that produce more return flux
    """Create a `StationaryModel` with narrower velocity bins in velocity regions with higher flux."""
    vg = VelocityGroups(vg)
    while True:
        print(f'Creating model with {vg['VGNumber']} velocity groups')
        m = self.stationary_floquet_model(mu, time, n_vector, nu, n_operator, vg, solver)
        print('Solving model')
        u = m.solve(mu_refine)
        o = m.reconstructed_flux_distribution(u, mu_refine).real.array.sum(['Time', 'Transition']).squeeze().data
        o /= np.sum(o)
        mask = o > max_weight
        if not np.sum(mask):
            break
        vg = vg.subdivide(mask, subdivisions)
    return self.stationary_floquet_model(mu, time, n_vector, nu, n_operator, vg, solver), u
